"""
Conversion data for all supported units
"""

from . import siprefix
from . import volume_consts
from . import mass_consts
from . import length_consts
from . import data_consts
from . import temperature_consts

_volume_conversions = {
    'oz': volume_consts.FL_OZ,    # Remove dots
    'gill': volume_consts.GILL,
    'cup': volume_consts.CUP,
    'pt': volume_consts.PT,
    'qt': volume_consts.QT,
    'gal': volume_consts.GAL,
    'tsp': volume_consts.TSP,
    'tbsp': volume_consts.TBSP,
    'dr': volume_consts.FL_DR,

    'imp oz': volume_consts.IMP_FL_OZ,
    'imp gill': volume_consts.IMP_GILL,
    'imp cup': volume_consts.IMP_CUP,
    'imp pt': volume_consts.IMP_PT,
    'imp qt': volume_consts.IMP_QT,
    'imp gal': volume_consts.IMP_GAL,
    'imp dr': volume_consts.IMP_FL_DR,

    'Yl': siprefix.YOTTA,
    'Zl': siprefix.ZETTA,
    'El': siprefix.EXA,
    'Pl': siprefix.PETA,
    'Tl': siprefix.TETRA,
    'Gl': siprefix.GIGA,
    'Ml': siprefix.MEGA,
    'kl': siprefix.KILO,
    'hl': siprefix.HECTO,
    'dal': siprefix.DECA,
    'l': siprefix.UNITY,
    'dl': siprefix.DECI,
    'cl': siprefix.CENTI,
    'ml': siprefix.MILLI,
    'ul': siprefix.MICRO,
    'nl': siprefix.NANO,
    'pl': siprefix.PICO,
    'fl': siprefix.FEMTO,
    'al': siprefix.ATTO,
    'zl': siprefix.ZEPTO,
    'yl': siprefix.YOCTO
}

_mass_conversions = {
    'oz': mass_consts.OZ,    # Remove dots
    'gr': mass_consts.GR,
    'dr': mass_consts.DR,
    'lb': mass_consts.LB,
    'cwt': mass_consts.CWT,

    'Yg': siprefix.YOTTA,
    'Zg': siprefix.ZETTA,
    'Eg': siprefix.EXA,
    'Pg': siprefix.PETA,
    'Tg': siprefix.TETRA,
    'Gg': siprefix.GIGA,
    'Mg': siprefix.MEGA,
    'kg': siprefix.KILO,
    'hg': siprefix.HECTO,
    'dag': siprefix.DECA,
    'g': siprefix.UNITY,
    'dg': siprefix.DECI,
    'cg': siprefix.CENTI,
    'mg': siprefix.MILLI,
    'ug': siprefix.MICRO,
    'ng': siprefix.NANO,
    'pg': siprefix.PICO,
    'fg': siprefix.FEMTO,
    'ag': siprefix.ATTO,
    'zg': siprefix.ZEPTO,
    'yg': siprefix.YOCTO,

    't': 1e3 * 1e3    # metric ton (ie, tonne = 1000kg)
}

_length_conversions = {
    'in': length_consts.IN,    # Remove dots
    'ft': length_consts.FT,
    'yd': length_consts.YD,
    'mi': length_consts.MI,
    'nmi': length_consts.NMI,

    'px': length_consts.PX,
    'pt': length_consts.PT,
    'pc': length_consts.PC,

    'Ym': siprefix.YOTTA,
    'Zm': siprefix.ZETTA,
    'Em': siprefix.EXA,
    'Pm': siprefix.PETA,
    'Tm': siprefix.TETRA,
    'Gm': siprefix.GIGA,
    'Mm': siprefix.MEGA,
    'km': siprefix.KILO,
    'hm': siprefix.HECTO,
    'dam': siprefix.DECA,
    'm': siprefix.UNITY,
    'dm': siprefix.DECI,
    'cm': siprefix.CENTI,
    'mm': siprefix.MILLI,
    'um': siprefix.MICRO,
    'nm': siprefix.NANO,
    'pm': siprefix.PICO,
    'fm': siprefix.FEMTO,
    'am': siprefix.ATTO,
    'zm': siprefix.ZEPTO,
    'ym': siprefix.YOCTO,
}

_data_conversions = {
    'YB': siprefix.YOTTA,
    'ZB': siprefix.ZETTA,
    'EB': siprefix.EXA,
    'PB': siprefix.PETA,
    'TB': siprefix.TETRA,
    'GB': siprefix.GIGA,
    'MB': siprefix.MEGA,
    'kB': siprefix.KILO,
    'B': siprefix.UNITY,

    'YiB': data_consts.YI,
    'ZiB': data_consts.ZI,
    'EiB': data_consts.EI,
    'PiB': data_consts.PI,
    'TiB': data_consts.TI,
    'GiB': data_consts.GI,
    'MiB': data_consts.MI,
    'KiB': data_consts.KI,

    'nibble': data_consts.NIBBLE,
    'b': data_consts.BIT,
}

_temperature_conversions = {
    'K': siprefix.UNITY,
    'C': temperature_consts.C,
    'F': temperature_consts.F,
    'Ro': temperature_consts.Ro,
    'Re': temperature_consts.Re,
    'De': temperature_consts.De,
    'Ra': temperature_consts.Ra,
    'N': temperature_consts.N,
}

CONVERSIONS = {
    'length': _length_conversions,
    'mass': _mass_conversions,
    'volume': _volume_conversions,
    'data': _data_conversions,
    'temperature': _temperature_conversions,
}

"""
Armstrong
"""



"""
SI prefixes for use by units
https://www.bipm.org/en/measurement-units/prefixes.html
"""

YOTTA = 1e24
ZETTA = 1e21
EXA = 1e18
PETA = 1e15
TETRA = 1e12
GIGA = 1e9
MEGA = 1e6
KILO = 1e3
HECTO = 1e2
DECA = 1e1
DECI = 1e-1
CENTI = 1e-2
MILLI = 1e-3
MICRO = 1e-6
NANO = 1e-9
PICO = 1e-12
FEMTO = 1e-15
ATTO = 1e-18
ZEPTO = 1e-21
YOCTO = 1e-24
UNITY = 1   # for reference value



"""
Temperature constants

Base unit: Kelvin
"""

from fractions import Fraction

# Temperature (in Kelvin) (K)
K = 1   # Kelvin
C = {"to": lambda x: x + 273.15, "from": lambda x: x - 273.15}  # Celsius
F = {   # Fahrenheit
    "to": lambda x: (x + 459.67) * Fraction(5 / 9),
    "from": lambda x: (x * Fraction(9 / 5)) - 459.67,
}
Ro = {  # Rømer
    "to": lambda x: (x - 7.5) * Fraction(40/21) + 273.15,
    "from": lambda x: (x - 273.15) * Fraction(21/40) + 7.5
}
Re = {  # Réaumur
    "to": lambda x: x * Fraction(5/4) + 273.15,
    "from": lambda x: (x - 273.15) * Fraction(4/5)
}
De = {  # Delisle
    "to": lambda x: 373.15 - (x * Fraction(2/3)),
    "from": lambda x: (373.15 - x) * Fraction(3/2)
}
Ra = {  # Rankine
    "to": lambda x: x * Fraction(5/9),
    "from": lambda x: x * Fraction(9/5)
}
N = {   # Newton
    "to": lambda x: (x * Fraction(100/33)) + 273.15,
    "from": lambda x: (x - 273.15) * Fraction(33/100)
}



"""
Data constants

Base unit: Byte
"""

# In bytes
BIT = 0.125     # 1/8
NIBBLE = 0.5    # 4/8

# IEC units
KI = 2**10  # kibi
MI = 2**20  # mebi
GI = 2**30  # gibi
TI = 2**40  # tebi
PI = 2**50  # pebi
EI = 2**60  # exbi
ZI = 2**70  # zebi
YI = 2**80  # yobi



"""
punits
"""

from typing import List, Optional
from fractions import Fraction

from punits.consts import LABELS, CONVERSIONS


def find_unit_code(measure: str,
                   unit_name: str) -> str:
    """
    Finds unit code for further processing

    Raises UnknownUnitException if the unit is unknown
    """
    unit_name = unit_name.replace(".", "")
    for label in LABELS[measure]:
        if (unit_name == label) or (unit_name in LABELS[measure][label]):
            # if label == 'NO_SHORT_CODE':
            #    return unit_name
            return label
    raise ValueError(f"Gee.. I don't know about '{unit_name}'..")


def to_from_base(measure: str,
                 unit_code: str,
                 to_from: str,
                 value: float,
                 params: dict) -> float:
    """
    measure: The type of units being converted
    unit_code: Code of unit
    to_from: Value of "to" indicates whether conversion
    is to take place from specified unit to base unit.
    Otherwise conversion is done from base unit to given
    unit.
     - "to": convert to base
     - "from": convert from base
    value: Value to be converted
    params: Any special parameters that will be needed
    for conversion. Like dpi for conversion involving
    pixels

    Converts a unit to or from the base unit of the
    unit's measure.

    Returns equivalent base unit value when to_from is
    "to" and equivalent unit value otherwise.
    """
    result: float
    conv = CONVERSIONS[measure][unit_code]
    if isinstance(conv, (int, float, Fraction)):
        # Linear units
        if to_from == 'to':
            # Convert to base unit
            result = value * conv
        else:
            # Convert from base unit
            result = value / conv
    else:
        # Non-linear units
        if params:
            result = conv[to_from](value, params)
        else:
            result = conv[to_from](value)
    return result


def punits(measure: str,
           src_unit: str,
           target_unit: str,
           values: List[float],
           params: Optional[dict] = None) -> List[float]:
    """
    measure: The type of units being converted
    src_unit: code of unit from which conversion
    is to take place
    target_unit: code of unit to which conversion
    is to take place
    values: List of values to be converted
    params: Any special parameters that will be needed
    for conversion. Like dpi for conversion involving
    pixels

    Returns converted values as a list
    """
    if params is None:
        params = {}

    results = []

    if ('dpi' not in params
            and (src_unit == 'px' or target_unit == 'px')):
        raise ValueError("Missing parameter: dpi")

    for value in values:
        base_value = to_from_base(measure, src_unit, "to", value, params)
        target_value = to_from_base(measure, target_unit,
                                    "from", base_value, params)
        results.append(target_value)
    return results


def get_factor(measure: str,
               src_unit_code: str,
               target_unit_code: str) -> Optional[float]:
    """
    src_unit: code of unit from which conversion is to take place
    target_unit: code of unit to which conversion is to take place
    conv_dict: CONVERSIONS[measure]

    Returns the factor of source unit by which target unit grows
    if source-target unit pair has a linear relationship, otherwise
    None is returned.

    To be used when --verbose option is used.
    """
    if(isinstance(CONVERSIONS[measure][src_unit_code], (int, float, Fraction))
       and isinstance(CONVERSIONS[measure][target_unit_code],
                      (int, float, Fraction))):
        target_factor = CONVERSIONS[measure][target_unit_code]
        src_factor = CONVERSIONS[measure][src_unit_code]
        factor = target_factor / src_factor
        return factor
    return None



"""
__main__.py for punits
"""

__version__ = "0.0.2"
__author__ = "Julin S"

import sys
import punits.app

if __name__ == "__main__":
    parser = punits.app.create_parser()
    args = parser.parse_args()
    sys.exit(punits.app.main(args))



"""
Mass constants

Base unit: Gram
"""

from fractions import Fraction

# US customary (mass in g) (avoirdupois - avdp)
OZ = Fraction(28.349523125)  # ounce
GR = Fraction(2/875) * OZ    # grain (16/7000)
DR = Fraction(1/16) * OZ     # dram
LB = 16 * OZ                 # pound
CWT = 100 * 16 * OZ           # US hundredweight(=100 lb)



"""
Command line interface for punits
"""

__version__ = "0.2"
__author__ = "Julin S"

import argparse

import punits

SUPPORTED_MEASURES = ['mass', 'length', 'volume', 'data', 'temperature']


def create_parser() -> argparse.ArgumentParser:
    """
    Create the parser for the cli.

    Returns the parser object.
    """
    parser = argparse.ArgumentParser(
        prog="punits",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description="A punits description for argparse"
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + __version__
    )

    # Type conversion is automatic when `type` arg is mentioned
    parser.add_argument('measure', choices=SUPPORTED_MEASURES)
    parser.add_argument('src_unit')
    parser.add_argument('target_unit')
    parser.add_argument('values', nargs='+', type=float)
    parser.add_argument('-p', '--precision', default=2, type=int)
    parser.add_argument('-v', '--verbose', action="store_true")

    # Relevant only for length
    parser.add_argument('--dpi', type=int,
                        help="Use when converting to or from px")
    return parser


def main(args: argparse.Namespace) -> int:
    """
    The main function for the CLI.

    Returns non-zero on error and zero on successful operation.
    """
    try:
        src_unit = punits.find_unit_code(args.measure, args.src_unit)
        target_unit = punits.find_unit_code(args.measure, args.target_unit)

        params = {}
        if args.dpi is not None:
            params['dpi'] = args.dpi

        results = punits.punits(args.measure, src_unit,
                                target_unit, args.values, params)
        str_results = [
            f"{round(result, args.precision):g}" for result in results]
        out_str = ' '.join(str_results)
        print(out_str)

        if args.verbose:
            factor = punits.get_factor(args.measure, src_unit, target_unit)
            if factor is not None:
                # Linear relationship between units
                print(f"\n1 {target_unit} = {factor:g} {src_unit} (approx)")
    except ValueError as ve:
        print(ve.args[0])
        return 1
    return 0



"""
Alternate names for all supported units
"""

from typing import Dict, List

_mass_labels: Dict[str, List[str]] = {
    # US customary
    'oz': ['ounce', 'ounces'],
    'gr': ['grain', 'grains'],
    'dr': ['dram', 'drams'],
    'lb': ['pound', 'pounds', 'lbs'],
    'cwt': ['hundredweight'],

    't': ['tonne', 'tonnes'],    # Same as Mg

    # SI
    'g': ['gram', 'grams'],
    'Yg': ['yottagram'],
    'Zg': ['zettagram'],
    'Eg': ['exagram'],
    'Pg': ['petagram'],
    'Tg': ['tetragram'],
    'Gg': ['gigagram'],
    'Mg': ['megagram'],
    'kg': ['kilogram'],
    'hg': ['hectogram'],
    'dag': ['decagram'],
    'dg': ['decigram'],
    'cg': ['centigram'],
    'mg': ['milligram'],
    'ug': ['microgram', 'μg'],
    'ng': ['nanogram'],
    'pg': ['picogram'],
    'fg': ['femtogram'],
    'ag': ['attogram'],
    'zg': ['zeptogram'],
    'yg': ['yoctogram'],
}

_volume_labels: Dict[str, List[str]] = {
    # US customary
    'oz': ['fluid ounce', 'ounce',
           'ounces', 'fl oz', 'oz fl'],    # Remove dots
    'gill': ['gill'],
    'cup': ['cup', 'cups'],
    'pt': ['pint'],
    'qt': ['quarter'],
    'gal': ['gallon', 'gallons'],
    'tsp': ['teaspoon'],
    'tbsp': ['tablespoon'],
    'dr': ['fluid dram', 'dram', 'drachm',
           'fluid drachm', 'fluidram', 'fluidrachm'],

    # Imperial
    'imp oz': ['imperial fluid ounce', 'imp_oz', 'imp fl oz', 'imp oz fl'],
    'imp gill': ['imperial gill', 'imp_gill'],
    'imp cup': ['imperial cup', 'imp_cup', 'imp cups'],
    'imp pt': ['imperial pint', 'imp_pt', 'imp pint'],
    'imp qt': ['imperial quarter', 'imp_qt', 'imp quarter'],
    'imp gal': ['imperial gallon', 'imp_gal',
                'imp gallon', 'imp gallons'],
    'imp dr': ['imperial fluid dram', 'imp_dr', 'imp dram',
               'imp drachm', 'imp fluid dram', 'imp fluid drachm',
               'imp fluidram', 'imp fluidrachm'],

    # SI
    'l': ['litre', 'litres'],
    'Yl': ['yottalitre'],
    'Zl': ['zettalitre'],
    'El': ['exalitre'],
    'Pl': ['petalitre'],
    'Tl': ['tetralitre'],
    'Gl': ['gigalitre'],
    'Ml': ['megalitre'],
    'kl': ['kilolitre'],
    'hl': ['hectolitre'],
    'dal': ['decalitre'],
    'dl': ['decilitre'],
    'cl': ['centilitre'],
    'ml': ['millilitre'],
    'ul': ['microlitre', 'μl'],
    'nl': ['nanolitre'],
    'pl': ['picolitre'],
    'fl': ['femtolitre'],
    'al': ['attolitre'],
    'zl': ['zeptolitre'],
    'yl': ['yoctolitre'],
}

_length_labels: Dict[str, List[str]] = {
    # US customary
    'in': ['inch', 'inches', '"'],
    'ft': ['feet', 'foot', "'"],
    'yd': ['yard', 'yards'],
    'mi': ['mile', 'miles'],
    'nmi': ['nautical mile', 'nautical miles'],

    # Web
    'px': ['pixel', 'pixels'],
    'pt': ['point', 'points'],
    'pc': ['pica', 'picas'],

    # SI
    'm': ['metre', 'metres',    # Commonwealth spelling
          'meter', 'meters'],   # US spelling
    'Ym': ['yottametre'],
    'Zm': ['zettametre'],
    'Em': ['exametre'],
    'Pm': ['petametre'],
    'Tm': ['tetrametre'],
    'Gm': ['gigametre'],
    'Mm': ['megametre'],
    'km': ['kilometre'],
    'hm': ['hectometre'],
    'dam': ['decametre'],
    'dm': ['decimetre'],
    'cm': ['centimetre'],
    'mm': ['millimetre'],
    'um': ['micrometre', 'μm'],
    'nm': ['nanometre'],
    'pm': ['picometre'],
    'fm': ['femtometre'],
    'am': ['attometre'],
    'zm': ['zeptometre'],
    'ym': ['yoctometre'],
}

_data_labels: Dict[str, List[str]] = {
    # Byte
    'B': ['byte', 'bytes'],
    'YB': ['yottabyte'],
    'ZB': ['zettabyte'],
    'EB': ['exabyte'],
    'PB': ['petabyte'],
    'TB': ['tetrabyte'],
    'GB': ['gigabyte'],
    'MB': ['megabyte'],
    'kB': ['kilobyte'],
    'YiB': ['yobibyte'],
    'ZiB': ['zebibyte'],
    'EiB': ['exbibyte'],
    'PiB': ['pebibyte'],
    'TiB': ['tebibyte'],
    'GiB': ['gibibyte'],
    'MiB': ['mebibyte'],
    'KiB': ['kebibyte'],

    'nibble': ['nibble'],

    # bit
    'b': ['bit', 'bits'],
    'Yb': ['yottabit'],
    'Zb': ['zettabit'],
    'Eb': ['exabit'],
    'Pb': ['petabit'],
    'Tb': ['tetrabit'],
    'Gb': ['gigabit'],
    'Mb': ['megabit'],
    'kb': ['kilobit'],
    'Yib': ['yobibit'],
    'Zib': ['zebibit'],
    'Eib': ['exbibit'],
    'Pib': ['pebibit'],
    'Tib': ['tebibit'],
    'Gib': ['gibibit'],
    'Mib': ['mebibit'],
    'Kib': ['kebibit'],
}

_temperature_labels: Dict[str, List[str]] = {
    'K': ['Kelvin', 'kelvin'],
    'C': ['Celsius', '°C', 'centigrade'],
    'F': ['Fahrenheit', '°F'],
    'N': ['Newton', 'N', '°N'],
    'Ro': ['Rømer', 'Rø', '°Rø', 'Ro',
           '°Ro', 'Romer'],  # 'R' is ignored to avoid conflict
    'Re': ['Réaumur', 'Ré', '°Re', '°r', 'r', 'Reaumur'],
    'De': ['Delisle', 'D', '°D', 'De', '°De'],
    'Ra': ['Rankine', '°Ra', 'Ra'],   # 'R' is ignored to avoid conflict
}

LABELS: Dict[str, Dict[str, List[str]]] = {
    'length': _length_labels,
    'mass': _mass_labels,
    'volume': _volume_labels,
    'data': _data_labels,
    'temperature': _temperature_labels,
}



"""
Length constants

Base unit: Metre
"""

from fractions import Fraction

# Length (US) (in m)
IN = 0.0254            # inch
FT = 12 * IN           # feet
YD = 36 * IN           # yard
MI = 1760 * (36 * IN)  # mile = 1760 yards
NMI = 1852


# Web
# W3C CSS: 1in = 2.54cm = 25.4mm = 72pt = 6pc
# https://www.w3.org/Style/Examples/007/units.en.html
PT = Fraction(1/72) * IN  # 'point' in typography. For font size, etc.
PC = Fraction(1/6) * IN   # 'pica' in typography. For font size, etc.

# PX = 0.0254 when dpi=1
PX = {"to": lambda x, args: (x * IN) / args['dpi'],
      "from": lambda x, args: (x / IN) * args['dpi']}



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="punits",
    version="0.2",
    author="Julin S",
    author_email="julins@cannottellyou.yet",
    description="A simple utility to convert values between different units",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ju-sh/punits",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        # https://pypi.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
        "Development Status :: 2 - Pre-Alpha",
        "Natural Language :: English",
    ],
    project_urls={
        'Changelog': 'https://github.com/ju-sh/punits/blob/master/CHANGELOG.md',
        'Issue Tracker': 'https://github.com/ju-sh/punits/issues',
    },
    install_requires=[],
    python_requires='>=3.6',
)



"""
Volume constants

Base unit: Litre
"""

from fractions import Fraction

# US liquid/fluid (volume in L) customary
# Note: US food labeling nutrition => oz = 30mL
FL_OZ = Fraction(0.0295735295625)  # ounce
GILL = 4 * FL_OZ                   # gill, teacup
CUP = 8 * FL_OZ                    # cup
PT = 16 * FL_OZ                    # pint
QT = 32 * FL_OZ                    # quart
GAL = 128 * FL_OZ                  # gallon
TBSP = Fraction(1/2) * FL_OZ       # tablespoon
TSP = Fraction(1/6) * FL_OZ        # teaspoon
FL_DR = Fraction(1/8) * FL_OZ      # fluid dram

# Imperial
IMP_FL_OZ = Fraction(0.0284130625)      # ounce
IMP_GILL = 5 * IMP_FL_OZ                # gill,teacup
IMP_CUP = 10 * IMP_FL_OZ                # cup
IMP_PT = 20 * IMP_FL_OZ                 # pint
IMP_QT = 40 * IMP_FL_OZ                 # quart
IMP_GAL = 160 * IMP_FL_OZ               # gallon
IMP_FL_DR = Fraction(1/8) * IMP_FL_OZ   # imperial fluid dram
